-[program "t2c.c" ]

-[headers, begin] !stdlib !string !stdio !errno !ctype -[end]

-[import strbuf library in private scope]
-[import proc library]

-[private] typedef struct line
{	struct line *prev, *next;
	int kind;
	char *data;
	char *fn;
	int lno;
} line_t;

-[private] typedef struct part
{
	struct part *prev, *next;
	int kind;
	char *target;
	line_t *lines;
	char *fn;
	int lno;
} part_t;

-[private] typedef struct block {
	struct block *prev, *next;
	char *name;
	line_t *lines;
	int busy;
} block_t;  // also used for files, where name==file name

line_t *read_file(char *fn)
{
	line_t *L;
	FILE *f;
	char g[2000];
	int lno;
	L= new_line(0,0);
	f= fopen(fn, "r");
	if (!f) exit(fprintf(stderr,"can't open %s: %s\n", fn,
		strerror(errno)));
	lno= 1;
	while(fgets(g, 1999,f))
	{
		line_t *K;
		if (strchr("+>:",g[0]))
			K= new_line(g[0], g+1);
		else
			K= new_line('T', g);
		K->fn= fn;
		K->lno= lno++;
		append_one(L, K);
	}
	fclose(f);
	return L;
}

-[private] line_t *new_line(int kind, char *data)
{
	line_t *L;
	if (kind!='T') normalize(data);
	-[allocate object L]
	L->kind= kind;
	if (data) L->data= strdup(data);
	else L->data= 0;
	L->prev= L->next= L;
	return L;
}

-[private] void normalize(char *txt)
{
	int i,j;
	int last;
	if (!txt) return ;
	for(i=0;txt[i];i++)
		if (txt[i]<=32 || txt[i]==127) txt[i]= ' ';

	last= ' ';
	j= 0;
	for(i=0;txt[i];i++) 
	   if (txt[i]!=' ' || last!=' ')
		last= txt[j++]= txt[i]; 
	txt[j]= 0;
	if (j>0 && txt[j-1]==' ') txt[j-1]= 0;
}

-[private] part_t *to_parts(line_t *lines)
{
	part_t *parts;
	-[allocate object parts]
	parts->prev= parts->next= parts;
	while(lines->next!=lines)
	{
		line_t *H;
		line_t *B;
		part_t *P;
		H= remove_one(lines);
		if (H->kind=='T' || H->kind==':')
			exit(fprintf(stderr,"%s: %d: expecting a "
				"redirection command\n", H->fn, H->lno));
		B= remove_body(lines);
		-[ allocate object P ]
		P->target= H->data;
		P->kind= H->kind;
		P->lines= B;
		P->fn= H->fn;
		P->lno= H->lno;

		P->prev= parts->prev;
		P->next= parts;
		P->prev->next= P;
		parts->prev= P;
	}
	return parts;
}

-[private] void do_prev_reloc(part_t *parts)
{
	part_t *P, *PP;
	for(P=parts->prev;P!=parts->next;P=PP)
	{
		part_t *T;
		PP= P->prev;
		if (P->kind!='+' || strcmp(P->target,"PREV")) continue;
		T= PP->prev;
		if (T==parts)
			exit(fprintf(stderr,"%s: %d: PREV directive "
			  "refers to past the beginning of the input file\n",
				P->fn, P->lno));
		append_multi(T->lines, P->lines);
		P->next->prev= PP;
		PP->next= P->next;
		free(P->target);
		free(P);
	}
}

-[private] void do_reloc(part_t *parts)
{
	part_t *P, *N;
	for(P=parts->next;P!=parts;P=N)
	{
		block_t *B;
		N= P->next;
		if (P->kind=='>') {
			P->prev->next= P->next;
			P->next->prev= P->prev;

			part_t *F;
			for(F=files->next;F!=files;F=F->next)
			  if (!strcmp(F->target, P->target))
			  {
				append_multi(F->lines,P->lines);
				break;
			  }
			if (F==files) {
				P->prev= files->prev;
				P->next= files;
				files->prev->next= P;
				files->prev= P;
			}
			continue;
		}
		B= find_block(P->target);
		append_multi(B->lines, P->lines);
		P->prev->next= P->next;
		P->next->prev= P->prev;
		free(P->target);
		free(P);
	}
}

-[private] int one(int ignore) { return 1; }

-[private] int compare_file(char *fn, char *B)
{
	FILE *f;
	int R,i;
	char g[2000];
	f= fopen(fn,"r");
	if (!f) return 1;
	while(*B) {
	    R= fread(g, 1, sizeof(g),f);
	    if (R<=0) return one(fclose(f));
	    for(i=0;i<R;i++)
		if (B[i]!=g[i]) return one(fclose(f));
	    B+= R;
	}
	R= fread(g, 1, sizeof(g),f);
	if (R>0) return one(fclose(f));
	fclose(f);
	return 0;
}

-[private] line_t *shallow_copy(line_t *L)
{
  line_t *C;
  C= malloc(sizeof(*L));
  *C= *L;
  C->prev= C->next= C;
  return C;
}

-[private] void collect_lines(line_t *dst, line_t *src)
{
   line_t *L;
   for(L=src->next;L!=src;L=L->next)
   {
     block_t *B;
     if (L->kind=='T')	
     {
        append_one(dst, shallow_copy(L));
        continue;
     }
     B= find_block(L->data);
     if (B->busy) {
	fprintf(stderr,"a recursion involving the following "
			"blocks was found:\n");
	for(B=blocks->next;B!=blocks;B=B->next)
	  if (B->busy)
		fprintf(stderr,"  %s\n", B->name);
	exit(1);
     }
     if (B->lines->next==B->lines)
	fprintf(stderr,"%s: %d: warning: writing out "
		  "empty block '%s'\n", L->fn, L->lno, L->data);
     B->busy= 1;
     collect_lines(dst, B->lines);
     B->busy= 0;
   }
}

-[private] void write_file(part_t *fb)
{
   line_t *lines;
   line_t *anchor, *body;
   char *B;
   lines= new_line(0,0);
   collect_lines(lines, fb->lines);
   while(find_innermost_pipe(lines, &anchor, &body))
   {
     char *text= collect_lines_text(body);
     char *args[200];
     char *otext; size_t otextl;
     char *etext; size_t etextl;
     int status;
     parse_args(anchor->data+1, (char**) args);
     if (proc_run(args[0], args, &status, text, strlen(text),
           &otext, &otextl, &etext, &etextl))
     {
        fprintf(stderr,"Child process %s failed.", args[0]);
        if (etextl) {
         fprintf(stderr,"The standard error stream was:\n");
         fwrite(etext, 1, etextl, stderr);
        }
        exit(1);
     }
     anchor->data= otext;
     anchor->kind= 'Z';
   }
   B= collect_lines_text(lines);

   if (force_output || compare_file(fb->target, B))
   {
      FILE *f= fopen(fb->target, "w");
      if (!f)
         exit(fprintf(stderr,"can not open %s: %s\n",
	    fb->target, strerror(errno)));
      if (fwrite(B, 1, strlen(B),f)!=strlen(B))
         exit(fprintf(stderr,"write error for %s: %s\n",
	    fb->target, strerror(ferror(f))));	
      free(B);
      fclose(f);
   }
}

-[private] int find_innermost_pipe(line_t *lines, line_t **A, line_t **B)
{
   line_t *begin=0;
   line_t *end=0;
   line_t *L;
   for(L=lines->next;L!=lines;L=L->next)
   {
     if (L->data[0]!='<') continue;
     if (L->kind=='Z') continue;
     if (isempty(L->data+1)) {
        end= L; break;
     } else {
        begin= L;
        end= 0;
     }
   }
   if (!begin && !end) return 0;
   if (begin && !end)
       exit(fprintf(stderr,"There is an unmatched pipe-open somewhere.\n"
          "Sorry I can't be more specific, this program was not supposed\n"
          "to give error messages. Have a nice day.\n"));
   if (end && !begin)
       exit(fprintf(stderr,"There is an unmatched pipe-close somewhere.\n"
          "Sorry I can't be more specific, this program was not supposed\n"
          "to give error messages. Have a nice day.\n"));
   line_t *body= new_line(0,0);
   while(begin->next!=end) {
     L= begin->next;
     remove_from_list(L);
     append_one(body, L);
   } 
   remove_from_list(end);
   *B= body;
   *A= begin;
   return 1;
}

-[private] void remove_from_list(line_t *L)
{
   L->prev->next= L->next;
   L->next->prev= L->prev;
}

-[private] char *collect_lines_text(line_t *lines)
{
   line_t *L; int S; char *R;
   S= 0;
   for(L=lines->next;L!=lines;L=L->next) S+= strlen(L->data);
   R= malloc(S+1);
   S= 0;
   for(L=lines->next;L!=lines;L=L->next) {
     memcpy(R+S, L->data, strlen(L->data));
     S+= strlen(L->data);
   }
   R[S]= 0;
   return R;
}

-[private] void write_file_old(part_t *fb)
{
	FILE *f;
        strbuf_t *SB;
 	char *B; 
        SB= strbuf_init();
	write_lines(SB,fb->lines);
	B= strbuf_collect(SB);
	if (force_output || compare_file(fb->target, B)) {
	   f= fopen(fb->target, "w");
	   if (!f)
              exit(fprintf(stderr,"can not open %s: %s\n",
		 fb->target, strerror(errno)));
	   if (fwrite(B, 1, strlen(B),f)!=strlen(B))
	      exit(fprintf(stderr,"write error for %s: %s\n",
		 fb->target, strerror(ferror(f))));	
	   free(B);
	   fclose(f);
	}
}

-[private] void write_lines(strbuf_t *f,line_t *lines)
{
	line_t *L;
	for(L=lines->next;L!=lines;L=L->next)
	{
		block_t *B;
		if (L->kind=='T')	
		{
			sbputs(f, L->data);
			continue;
		}
		B= find_block(L->data);
		if (B->busy) {
			fprintf(stderr,"a recursion involving the following "
				"blocks was found:\n");
			for(B=blocks->next;B!=blocks;B=B->next)
			  if (B->busy)
				fprintf(stderr,"  %s\n", B->name);
			exit(1);
		}
		if (B->lines->next==B->lines)
			fprintf(stderr,"%s: %d: warning: writing out "
			  "empty block '%s'\n", L->fn, L->lno, L->data);
		B->busy= 1;
		write_lines(f, B->lines);
		B->busy= 0;
	}
}

-[private] block_t *find_block(char *name)
{
	block_t *B;
	for(B=blocks->next;B!=blocks;B=B->next)
		if (!strcmp(B->name, name))
			return B;
	-[allocate object B]
	B->busy= 0;
	B->name= strdup(name);
	B->lines= new_line(0,0);
	B->prev= blocks->prev;
	B->next= blocks;
	blocks->prev->next= B;
	blocks->prev= B;
	return B;
}

-[private] block_t *blocks;
-[private] part_t *files;

-[private] int force_output;
-[private] void shift_args(int *ac,char **av, int k)
{
   int i;
   for(i=k;i<*ac-1;i++)
	av[i]= av[i+1];
   (*ac)--;
}


int main(int argc,char **argv)
{
	int i;
	part_t *F;
	for(i=1;i<argc;i++) {
	   if (!strcmp(argv[i], "-f")) {
	     force_output=1;
	     shift_args(&argc,argv,i);
	     i--;
	   }
	}
	-[allocate object blocks] blocks->prev= blocks->next= blocks;
	-[allocate object files] files->prev= files->next= files;
	for(i=1;i<argc;i++) {
		part_t *parts;
		parts= to_parts(read_file(argv[i]));
		do_prev_reloc(parts);
		do_reloc(parts);
	}

#if 0
	{ block_t *B;
	for(B=blocks->next;B!=blocks;B=B->next) {
		printf("BLOCK '%s'\n", B->name);
		line_t *L;
		for(L=B->lines->next;L!=B->lines;L=L->next)
			printf("   %s", L->data);
		printf("------\n");
	}}
#endif
	// print_blocks_list(blocks);
        sort_blocks(blocks);
        // print_blocks_list(blocks);
	unify_blocks(blocks);
        // print_blocks_list(blocks);

        remove_numeric_suffixes(blocks);

	for(F=files->next;F!=files;F=F->next)
		write_file(F);
	return 0;
}


-[private] void append_multi(line_t *dst, line_t *src)
{
	line_t *dH, *dT, *sH, *sT;

	dH= dst;
	dT= dst->prev;

	sH= src->next;
	sT= src->prev;
/*    dT              -> dH
          sH -> ... sT        */
	dT->next= sH;
	sH->prev= dT;
	sT->next= dH;
	dH->prev= sT;
}

-[private] void prepend_multi(line_t *dst, line_t *src)
{
	line_t *dH, *dF, *sH, *sT;

	dH= dst;
	dF= dst->next;

	sH= src->next;
	sT= src->prev;
/*    dH          ->  dF
        sH .. sT            */
	dH->next= sH;
	sH->prev= dH;
	sT->next= dF;
	dF->prev= sT;
}

-[private] line_t *remove_one(line_t* lines)
{
	line_t *L;
	L= lines->next;
	L->prev->next= L->next;
	L->next->prev= L->prev;
	return L;
}

-[private] line_t *remove_body(line_t *lines)
{
	line_t *L, *K;
	L= new_line(0,0);

	while(lines->next!=lines)
	{
		K= lines->next;
		if (K->kind!='T' && K->kind!=':') break;
		remove_one(lines);
		append_one(L,K);
	}
	return L;
}

-[private] void append_one(line_t *L,line_t *K)
{
	K->prev= L->prev;
	K->next= L;
	L->prev->next= K;
	L->prev= K;
}

-[unused symbol sbprintf]


-[private] void parse_args(char *str, char **args)
{
  char *b;
  char *R;
again:
  while(is_real_space(*str)) str++;
  if (!str[0]) { *args=0; return ; }
  b= str;
  if (str[0]=='\'') {
    b++;
    str++;
    while(str[0] && str[0]!='\'') str++;
    if (!str[0])
       exit(fprintf(stderr,"There is an unmatched quote somewhere.\n"
          "Sorry I can't be more specific, this program was not supposed\n"
          "to give error messages. Have a nice day.\n"));
  } else {
    while(!is_real_space(*str)) str++;
  }
  R= malloc(str-b+1);
  memcpy(R, b, str-b);
  R[str-b]= 0;
  *args= R;
  args++;
  if (str[0]=='\'') str++;
  goto again;
}

-[private] int isempty(char *S)
{
  while(*S) if (!is_real_space(*S)) return 0; else S++;
  return 1;
}

-[private] int is_real_space(unsigned int A)
{
   return A && (A<=32 || A==127);
}

-[unused symbol write_file_old]
-[unused symbol code_to_str]


/* Now, we want to implement the numbered blocks feature. For this,
   I need to first sort the blocks and then combine them if they 
   have the same name (but different numbers). 

   In order to do this, I need to compare block names in a special
   way. For non-digit characters, we can simply compare the byte
   values. For digits, we need to compare the numbers represented
   by the following sequences of digits. This shall be done only if
   the digits are at the end of the string. Also, a string without
   a number suffix shall be sorted after the ones with digits. */

-[private] int compare_names(char *A, char *B)
{
   char *dA, *dB;
   int nA, nB, lA, lB;
   int R;

   if (!A || !A[0]) { if (B && B[0]) return -1; return 0; }
   if (!B || !B[0]) return 1; 

   dA= get_digits(A, &nA);
   dB= get_digits(B, &nB);

   /* here, I`m sorting stuff first based on their LENGTH instead of
   the byte values. This is acceptable since we don`t really need
   a lexical order but just some consistent order. After that, we sort
   on the lexical order. */

   lA= dA-A; lB= dB-B;
   R= lA-lB;
   if (R) return R;
   R= memcmp(A,B, lA);
   if (R) return R;

   if (nA==-1) { if (nB==-1) return 0; return 1; }
   if (nB==-1) { return -1; }
   return nA-nB;
}

/* When doing the unification, we will also need a function which determines
   whether two consecutive names are the same, ignoring the number suffix. */

-[private] int same_name(char *A, char *B)
{
   char *dA, *dB;
   int nA, nB, lA, lB;

   if (!A || !A[0]) { if (B && B[0]) return 0; return 1; }
   if (!B || !B[0]) return 0; 

   dA= get_digits(A, &nA);
   dB= get_digits(B, &nB);

   lA= dA-A; lB= dB-B;
   if (lA!=lB) return 0;
   return !memcmp(A,B, lA);
}

/* Below is the sorting function, a variant on quicksort. 
   Instead of swapping stuff, we insert elements into their
   respective lists and recurse on those. */

-[private] void sort_blocks(block_t *BB)
{
  block_t *key; 
  block_t head, mid, tail;
  block_t *B;

  if (BB->next==BB->prev) return ;

  head.prev= head.next= &head;
  mid.prev= mid.next= &mid;
  tail.prev= tail.next= &tail;

  key= blocks_remove_first(BB);
  blocks_append(&mid, key); 
  while((B=blocks_remove_first(BB))) {
    block_t *target;
    int R;
    R= compare_names(B->name,key->name);
    if (R<0) target= &head; else if (R>0) target= &tail; else target= &mid;
    blocks_append(target,B);
  }
  sort_blocks(&head); 
  sort_blocks(&tail); 
  blocklist_append_d(BB, &head);
  blocklist_append_d(BB, &mid);
  blocklist_append_d(BB, &tail);
}

-[private] block_t *blocks_remove_first(block_t *head)
{
  block_t *F;
  if (head->next==head) return 0;
  F= head->next;
  F->next->prev= head;
  head->next= F->next;
  F->prev= F->next= F;
  return F;
}

-[private] void blocks_append(block_t *head, block_t *B)
{
  B->prev= head->prev;
  B->next= head;
  head->prev->next= B;
  head->prev= B;
}

-[private] void blocklist_append_d(block_t *target, block_t *src)
{
  block_t *sF, *sL;
  if (src->next==src) return ;
  sF= src->next;
  sL= src->prev;

  sF->prev= target->prev;
  target->prev->next= sF;
 
  sL->next= target;
  target->prev= sL;
}

/* Now that we have the sorting stuff, we can easily unify them. But
   first, let`s have something that prints out the current list of
   blocks, so that I can debug this. */

-[private] void print_blocks_list(block_t *L)
{
  block_t *B;
  printf("BLOCKS:\n");
  for(B=L->next;B!=L;B=B->next)
     printf(" %s\n", B->name);
  printf("------\n");
}

/* Another utility function is the get_digits function. It returns the
   end of non-digit prefix and the integer representation of the digit
   suffix. -1 is returned as the number if the string doesn`t end with
   digits. */
-[private] char *get_digits(char *str, int *number)
{
  char *K;
  if (!str || !str[0]) goto nodigits;
  K= str+strlen(str)-1;
  while(isspace(*K)) K--;
  if (*K==0) goto nodigits;
  if (!isdigit(*K)) goto nodigits;
  while(isdigit(*K)) K--;
  *number= atoi(K+1);
  return K;
nodigits:
  *number= -1; return str+strlen(str); 
}

/* So far, I`ve debugged the thing. I hope it works in the general case.
   Anyway, now I can unify the whole thing and be done with it. */

-[private] void unify_blocks(block_t *BB)
{
  block_t *B,*N;
  for(B=BB->next;B->next!=BB;B=N)
  {
    N= B->next;
    if (same_name(B->name,N->name))
    {
       prepend_multi(N->lines, B->lines);
       B->prev->next= B->next;
       B->next->prev= B->prev;
       // FIXME: maybe free the now-useless block_t ?
    }
  }
}

/* now everything seems to be OK. Let`s put this in the wild. */

-[ unused symbol print_blocks_list ]

/* I have discovered a bug. When you have

 + Types 100
 + Types 200

 but not

 + Types

the whole thing disappears unless somebody explicitly includes 

 : Types 200

This is to be expected, but is quite wrong. So, I should remove
the number suffixes from block names when I`m done merging. */

-[private] void remove_numeric_suffix(char *title)
{
  char *S;
  int nr;
  S= get_digits(title, &nr);
  if (nr==-1) return ;
  while(isspace(*S)) { *S=0; S--; } 
}

-[private] void remove_numeric_suffixes(block_t *blocks)
{
  block_t *B;
  for(B= blocks->next;B!=blocks;B=B->next)
   remove_numeric_suffix(B->name);
}

/* This does it. Now :Types will bring in + Types 200 etc.
  However, : Types 200 will not work. */

